import os
import glob
import numpy as np

from ..util.load_dataset import load_set_xy
from ..util.split_validation import split_validation


def get_musdb(root_path):
    # Not yet implemented
    pass


def get_musdbhq(root_path):
    """
    Script to load musdbhq
    :param root_path: the root directory of MUSDB18HQ (containing train and test directories)
    :return: A dictionary {"train", "val", "test"} containing lists of dictionaries {"bass", "drums", "mixture", "other", "vocals"} containing the absolute path to respective track
    """
    subsets = dict()
    instruments = ["bass", "drums", "mixture", "other", "vocals"]

    for subset in ["train", "test"]:
        subsets[subset] = load_set_xy(
            os.path.join(root_path, subset), instruments=instruments
        )

    subsets["val"], subsets["train"] = split_validation(subsets["train"])
    return subsets



import torch
import torch.nn as nn

from ..model.crop import centre_crop
from ..model.resample import Resample1d
from ..model.conv import ConvLayer


class UpsamplingBlock(nn.Module):
    def __init__(
        self,
        n_inputs,
        n_shortcut,
        n_outputs,
        kernel_size,
        stride,
        depth,
        conv_type,
        res,
    ):
        super(UpsamplingBlock, self).__init__()
        assert stride > 1

        # CONV 1 for UPSAMPLING
        if res == "fixed":
            self.upconv = Resample1d(n_inputs, 15, stride, transpose=True)
        else:
            self.upconv = ConvLayer(
                n_inputs, n_inputs, kernel_size, stride, conv_type, transpose=True
            )

        self.pre_shortcut_convs = nn.ModuleList(
            [ConvLayer(n_inputs, n_outputs, kernel_size, 1, conv_type)]
            + [
                ConvLayer(n_outputs, n_outputs, kernel_size, 1, conv_type)
                for _ in range(depth - 1)
            ]
        )

        # CONVS to combine high- with low-level information (from shortcut)
        self.post_shortcut_convs = nn.ModuleList(
            [ConvLayer(n_outputs + n_shortcut, n_outputs, kernel_size, 1, conv_type)]
            + [
                ConvLayer(n_outputs, n_outputs, kernel_size, 1, conv_type)
                for _ in range(depth - 1)
            ]
        )

    def forward(self, x, shortcut):
        # UPSAMPLE HIGH-LEVEL FEATURES
        upsampled = self.upconv(x)

        for conv in self.pre_shortcut_convs:
            upsampled = conv(upsampled)

        # Prepare shortcut connection
        combined = centre_crop(shortcut, upsampled)

        # Combine high- and low-level features
        combined = torch.cat([combined, centre_crop(upsampled, combined)], dim=1)
        for conv in self.post_shortcut_convs:
            combined = conv(combined)
        return combined

    def get_output_size(self, input_size):
        curr_size = self.upconv.get_output_size(input_size)

        # Upsampling convs
        for conv in self.pre_shortcut_convs:
            curr_size = conv.get_output_size(curr_size)

        # Combine convolutions
        for conv in self.post_shortcut_convs:
            curr_size = conv.get_output_size(curr_size)

        return curr_size


class DownsamplingBlock(nn.Module):
    def __init__(
        self,
        n_inputs,
        n_shortcut,
        n_outputs,
        kernel_size,
        stride,
        depth,
        conv_type,
        res,
    ):
        super(DownsamplingBlock, self).__init__()
        assert stride > 1

        self.kernel_size = kernel_size
        self.stride = stride

        # CONV 1
        self.pre_shortcut_convs = nn.ModuleList(
            [ConvLayer(n_inputs, n_shortcut, kernel_size, 1, conv_type)]
            + [
                ConvLayer(n_shortcut, n_shortcut, kernel_size, 1, conv_type)
                for _ in range(depth - 1)
            ]
        )

        self.post_shortcut_convs = nn.ModuleList(
            [ConvLayer(n_shortcut, n_outputs, kernel_size, 1, conv_type)]
            + [
                ConvLayer(n_outputs, n_outputs, kernel_size, 1, conv_type)
                for _ in range(depth - 1)
            ]
        )

        # CONV 2 with decimation
        if res == "fixed":
            self.downconv = Resample1d(
                n_outputs, 15, stride
            )  # Resampling with fixed-size sinc lowpass filter
        else:
            self.downconv = ConvLayer(
                n_outputs, n_outputs, kernel_size, stride, conv_type
            )

    def forward(self, x):
        # PREPARING SHORTCUT FEATURES
        shortcut = x
        for conv in self.pre_shortcut_convs:
            shortcut = conv(shortcut)

        # PREPARING FOR DOWNSAMPLING
        out = shortcut
        for conv in self.post_shortcut_convs:
            out = conv(out)

        # DOWNSAMPLING
        out = self.downconv(out)

        return out, shortcut

    def get_input_size(self, output_size):
        curr_size = self.downconv.get_input_size(output_size)

        for conv in reversed(self.post_shortcut_convs):
            curr_size = conv.get_input_size(curr_size)

        for conv in reversed(self.pre_shortcut_convs):
            curr_size = conv.get_input_size(curr_size)
        return curr_size


class Waveunet(nn.Module):
    def __init__(
        self,
        num_inputs,
        num_channels,
        num_outputs,
        instruments,
        kernel_size,
        target_output_size,
        conv_type,
        res,
        separate=False,
        depth=1,
        strides=2,
    ):
        super(Waveunet, self).__init__()

        self.num_levels = len(num_channels)
        self.strides = strides
        self.kernel_size = kernel_size
        self.num_inputs = num_inputs
        self.num_outputs = num_outputs
        self.depth = depth
        self.instruments = instruments
        self.separate = separate

        # Only odd filter kernels allowed
        assert kernel_size % 2 == 1

        self.waveunets = nn.ModuleDict()

        model_list = instruments if separate else ["ALL"]
        # Create a model for each source if we separate sources separately, otherwise only one (model_list=["ALL"])
        for instrument in model_list:
            module = nn.Module()

            module.downsampling_blocks = nn.ModuleList()
            module.upsampling_blocks = nn.ModuleList()

            for i in range(self.num_levels - 1):
                in_ch = num_inputs if i == 0 else num_channels[i]

                module.downsampling_blocks.append(
                    DownsamplingBlock(
                        in_ch,
                        num_channels[i],
                        num_channels[i + 1],
                        kernel_size,
                        strides,
                        depth,
                        conv_type,
                        res,
                    )
                )

            for i in range(0, self.num_levels - 1):
                module.upsampling_blocks.append(
                    UpsamplingBlock(
                        num_channels[-1 - i],
                        num_channels[-2 - i],
                        num_channels[-2 - i],
                        kernel_size,
                        strides,
                        depth,
                        conv_type,
                        res,
                    )
                )

            module.bottlenecks = nn.ModuleList(
                [
                    ConvLayer(
                        num_channels[-1], num_channels[-1], kernel_size, 1, conv_type
                    )
                    for _ in range(depth)
                ]
            )

            # Output conv
            outputs = num_outputs if separate else num_outputs * len(instruments)
            module.output_conv = nn.Conv1d(num_channels[0], outputs, 1)

            self.waveunets[instrument] = module

        self.set_output_size(target_output_size)

    def set_output_size(self, target_output_size):
        self.target_output_size = target_output_size

        self.input_size, self.output_size = self.check_padding(target_output_size)
        print(
            "Using valid convolutions with "
            + str(self.input_size)
            + " inputs and "
            + str(self.output_size)
            + " outputs"
        )

        assert (self.input_size - self.output_size) % 2 == 0
        self.shapes = {
            "output_start_frame": (self.input_size - self.output_size) // 2,
            "output_end_frame": (self.input_size - self.output_size) // 2
            + self.output_size,
            "output_frames": self.output_size,
            "input_frames": self.input_size,
        }

    def check_padding(self, target_output_size):
        # Ensure number of outputs covers a whole number of cycles so each output in the cycle is weighted equally during training
        bottleneck = 1

        while True:
            out = self.check_padding_for_bottleneck(bottleneck, target_output_size)
            if out is not False:
                return out
            bottleneck += 1

    def check_padding_for_bottleneck(self, bottleneck, target_output_size):
        module = self.waveunets[[k for k in self.waveunets.keys()][0]]
        try:
            curr_size = bottleneck
            for idx, block in enumerate(module.upsampling_blocks):
                curr_size = block.get_output_size(curr_size)
            output_size = curr_size

            # Bottleneck-Conv
            curr_size = bottleneck
            for block in reversed(module.bottlenecks):
                curr_size = block.get_input_size(curr_size)
            for idx, block in enumerate(reversed(module.downsampling_blocks)):
                curr_size = block.get_input_size(curr_size)

            assert output_size >= target_output_size
            return curr_size, output_size
        except AssertionError as e:
            return False

    def forward_module(self, x, module):
        """
        A forward pass through a single Wave-U-Net (multiple Wave-U-Nets might be used, one for each source)
        :param x: Input mix
        :param module: Network module to be used for prediction
        :return: Source estimates
        """
        shortcuts = []
        out = x

        # DOWNSAMPLING BLOCKS
        for block in module.downsampling_blocks:
            out, short = block(out)
            shortcuts.append(short)

        # BOTTLENECK CONVOLUTION
        for conv in module.bottlenecks:
            out = conv(out)

        # UPSAMPLING BLOCKS
        for idx, block in enumerate(module.upsampling_blocks):
            out = block(out, shortcuts[-1 - idx])

        # OUTPUT CONV
        out = module.output_conv(out)
        if not self.training:  # At test time clip predictions to valid amplitude range
            out = out.clamp(min=-1.0, max=1.0)
        return out

    def forward(self, x, inst=None):
        curr_input_size = x.shape[-1]
        assert (
            curr_input_size == self.input_size
        )  # User promises to feed the proper input himself, to get the pre-calculated (NOT the originally desired) output size

        if self.separate:
            return {inst: self.forward_module(x, self.waveunets[inst])}
        else:
            assert len(self.waveunets) == 1
            out = self.forward_module(x, self.waveunets["ALL"])

            out_dict = {}
            for idx, inst in enumerate(self.instruments):
                out_dict[inst] = out[
                    :, idx * self.num_outputs : (idx + 1) * self.num_outputs
                ]
            return out_dict



import numpy as np
import librosa


def load_audio(audio_path, sr=44100, mono=False):
    """ """
    data, sanity_sr = librosa.load(audio_path, sr=sr, mono=mono)

    if len(data.shape) == 1:
        data = data[np.newaxis, :]

    return data, sanity_sr



import os

import h5py
import numpy as np
from torch.utils.data import Dataset
from sortedcontainers import SortedList
from tqdm import tqdm

from ..util.load_audio import load_audio


class WaveunetShuffleDataset(Dataset):
    """ """

    def __init__(
        self,
        dataset,
        hdf_dir,
        filename,
        instruments,
        sr,
        channels,
        shapes,
        random_hops,
        audio_transform=None,
        in_memory=False,
    ):
        """
        :param dataset: (list of dictionaries)
        """
        super(WaveunetShuffleDataset, self).__init__()

        self.hdf_dataset = None
        self.hdf_dir = hdf_dir
        self.hdf_file = os.path.join(hdf_dir, filename + ".hdf5")

        self.random_hops = random_hops
        self.sr = sr
        self.channels = channels
        self.shapes = shapes
        self.audio_transform = audio_transform
        self.in_memory = in_memory
        self.instruments = instruments

        if not os.path.exists(self.hdf_file):
            if not os.path.exists(hdf_dir):
                os.makedirs(hdf_dir)

            with h5py.File(self.hdf_file, "w") as f:
                f.attrs["sr"] = sr
                f.attrs["channels"] = channels
                f.attrs["instruments"] = instruments

                print("Adding audio files to dataset...")
                for idx, track in enumerate(tqdm(dataset)):
                    grp = f.create_group(str(idx))
                    for stem in instruments:
                        audio_data, _ = load_audio(
                            track[stem], sr=self.sr, mono=(self.channels == 1)
                        )
                        grp.create_dataset(
                            stem,
                            shape=audio_data.shape,
                            dtype=audio_data.dtype,
                            data=audio_data,
                        )

                    grp.attrs["target_length"] = audio_data.shape[1]

        with h5py.File(self.hdf_file, "r") as f:
            if (
                f.attrs["sr"] != sr
                or f.attrs["channels"] != channels
                or list(f.attrs["instruments"]) != instruments
            ):
                raise ValueError(
                    "SR or channel or instruments not the same as the already existing hdf file"
                )

        with h5py.File(self.hdf_file, "r") as f:
            lengths = [
                f[str(song_idx)].attrs["target_length"] for song_idx in range(len(f))
            ]
            lengths = [(l // self.shapes["output_frames"]) + 1 for l in lengths]

        self.start_pos = SortedList(np.cumsum(lengths))
        self.length = self.start_pos[-1]

        self.inst_shuffle = {key: np.arange(self.length) for key in self.instruments}
        self.shuffle()

    def __getitem__(self, index):
        if self.hdf_dataset is None:
            driver = "core" if self.in_memory else None
            self.hdf_dataset = h5py.File(self.hdf_file, "r", driver=driver)

        targets = {}
        for stem in self.instruments:
            targets[stem] = self.getStem(stem, self.inst_shuffle[stem][index])

        if self.audio_transform is not None:
            targets = self.audio_transform(targets)

        mixture = np.sum(list(targets.values()), axis=0)
        mixture = np.clip(mixture, -1, 1)

        for key in targets:
            targets[key] = targets[key][
                :, self.shapes["output_start_frame"] : self.shapes["output_end_frame"]
            ]
        return mixture, targets

    def getStem(self, stem, index):
        audio_idx = self.start_pos.bisect_right(index)
        if audio_idx > 0:
            index = index - self.start_pos[audio_idx - 1]

        target_length = self.hdf_dataset[str(audio_idx)].attrs["target_length"]
        if self.random_hops:
            start_target_pos = np.random.randint(
                0, max(target_length - self.shapes["output_frames"] + 1, 1)
            )
        else:
            start_target_pos = index * self.shapes["output_frames"]

        start_pos = start_target_pos - self.shapes["output_start_frame"]
        if start_pos < 0:
            pad_front = abs(start_pos)
            start_pos = 0
        else:
            pad_front = 0

        end_pos = (
            start_target_pos
            - self.shapes["output_start_frame"]
            + self.shapes["input_frames"]
        )
        if end_pos > target_length:
            pad_back = end_pos - target_length
            end_pos = target_length
        else:
            pad_back = 0

        target = self.hdf_dataset[str(audio_idx)][stem][:, start_pos:end_pos].astype(
            np.float32
        )
        if pad_front > 0 or pad_back > 0:
            target = np.pad(
                target,
                [(0, 0), (pad_front, pad_back)],
                mode="constant",
                constant_values=0.0,
            )
        return target

    def shuffle(self):
        for key in self.inst_shuffle:
            np.random.shuffle(self.inst_shuffle[key])

    def __len__(self):
        return self.length



import argparse
import os

import data.utils
import model.utils as model_utils

from test import predict_song
from model.waveunet import Waveunet


def main(args):
    # MODEL
    num_features = (
        [args.features * i for i in range(1, args.levels + 1)]
        if args.feature_growth == "add"
        else [args.features * 2**i for i in range(0, args.levels)]
    )
    target_outputs = int(args.output_size * args.sr)
    model = Waveunet(
        args.channels,
        num_features,
        args.channels,
        args.instruments,
        kernel_size=args.kernel_size,
        target_output_size=target_outputs,
        depth=args.depth,
        strides=args.strides,
        conv_type=args.conv_type,
        res=args.res,
        separate=args.separate,
    )

    if args.cuda:
        model = model_utils.DataParallel(model)
        print("move model to gpu")
        model.cuda()

    print("Loading model from checkpoint " + str(args.load_model))
    state = model_utils.load_model(model, None, args.load_model, args.cuda)
    print("Step", state["step"])

    preds = predict_song(args, args.input, model)

    output_folder = os.path.dirname(args.input) if args.output is None else args.output
    for inst in preds.keys():
        data.utils.write_wav(
            os.path.join(
                output_folder, os.path.basename(args.input) + "_" + inst + ".wav"
            ),
            preds[inst],
            args.sr,
        )


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--instruments",
        type=str,
        nargs="+",
        default=["bass", "drums", "other", "vocals"],
        help='List of instruments to separate (default: "bass drums other vocals")',
    )
    parser.add_argument("--cuda", action="store_true", help="Use CUDA (default: False)")
    parser.add_argument(
        "--features", type=int, default=32, help="Number of feature channels per layer"
    )
    parser.add_argument(
        "--load_model",
        type=str,
        default="checkpoints/waveunet/model",
        help="Reload a previously trained model",
    )
    parser.add_argument("--batch_size", type=int, default=4, help="Batch size")
    parser.add_argument("--levels", type=int, default=6, help="Number of DS/US blocks")
    parser.add_argument(
        "--depth", type=int, default=1, help="Number of convs per block"
    )
    parser.add_argument("--sr", type=int, default=44100, help="Sampling rate")
    parser.add_argument(
        "--channels", type=int, default=2, help="Number of input audio channels"
    )
    parser.add_argument(
        "--kernel_size",
        type=int,
        default=5,
        help="Filter width of kernels. Has to be an odd number",
    )
    parser.add_argument(
        "--output_size", type=float, default=2.0, help="Output duration"
    )
    parser.add_argument("--strides", type=int, default=4, help="Strides in Waveunet")
    parser.add_argument(
        "--conv_type",
        type=str,
        default="gn",
        help="Type of convolution (normal, BN-normalised, GN-normalised): normal/bn/gn",
    )
    parser.add_argument(
        "--res",
        type=str,
        default="fixed",
        help="Resampling strategy: fixed sinc-based lowpass filtering or learned conv layer: fixed/learned",
    )
    parser.add_argument(
        "--separate",
        type=int,
        default=1,
        help="Train separate model for each source (1) or only one (0)",
    )
    parser.add_argument(
        "--feature_growth",
        type=str,
        default="double",
        help="How the features in each layer should grow, either (add) the initial number of features each time, or multiply by 2 (double)",
    )

    parser.add_argument(
        "--input",
        type=str,
        default=os.path.join("audio_examples", "Cristina Vane - So Easy", "mix.mp3"),
        help="Path to input mixture to be separated",
    )
    parser.add_argument(
        "--output",
        type=str,
        default=None,
        help="Output path (same folder as input path if not set)",
    )

    args = parser.parse_args()

    main(args)



import os
import glob
import numpy as np


def load_set_xy(path, instruments):
    """
    Loads the set in 0th axis configuration (a lot of directories, each with tracks with the instrument names)
    :param path: Path to the set (train set / test set, not the full dataset)
    :param instrument: List of instruments that will be loaded (the name of the track files must be exact)
    :return: List of dictionaries, each with key-instrument names and value-path to the track
    """
    tracks = glob.glob(os.path.join(path, "*"))
    samples = list()

    for track_folder in sorted(tracks):
        track = dict()
        for stem in instruments:
            audio_path = os.path.join(track_folder, stem + ".wav")
            track[stem] = audio_path
        samples.append(track)

    return samples


def load_set_yx(path, instruments):
    """
    Loads the set in 1st axis configuration (instrument length of directories, filled with a lot of tracks)
    :param path: Path to the set (train set / test set, not the full dataset)
    :param instrument: List of instruments that will be loaded (the name of the directories must be exact)
    :return: dictionary of Lists, each with key-instrument names and value-list of paths to the track
    """
    samples = dict()

    for track in instruments:
        samples[track] = sorted(glob.glob(os.pash.join(path, track, "*")))

    return samples


def yx_to_xy(yx_set, shuffle=True):
    """
    converts the return of load_set_yx to be the same configuration as the return of load_set_xy (minimum length will be used)
    :param yx_set: Return value of load_set_yx
    :param shuffle: shuffles the configuration
    :return: Return value of load_set_xy
    """
    subsets = list()
    min_length = min([len(i) for i in yx_set])

    if shuffle:
        for key in yx_set:
            yx_set[key] = np.shuffle(yx_set[key])

    for i in min_length:
        sample = {key: yx_set[key][i] for key in yx_set}
        subsets.append(sample)

    return subsets



import numpy as np


def shuffle_train_set(train_set, it):
    """
    Shuffles the train dataset
    :param train_set: A list containing dictionaries [{"bass": bass_path, ...}, {}, {}, ...]
    :param it: integer, used as seed (everytime the dataset needs to be reshuffled, it should be changed)
    :return: The same shape as the input, mixture removed, shuffled
    """
    np.random.seed(it)
    length = len(train_set)

    bass = np.arange(length)
    drums = np.arange(length)
    other = np.arange(length)
    vocals = np.arange(length)

    np.shuffle(bass)
    np.shuffle(drums)
    np.shuffle(other)
    np.shuffle(vocals)

    return [
        {"bass": bass[i], "drums": drums[i], "other": other[i], "vocals": vocals[i]}
        for i in length
    ]



import numpy as np


def split_validation(trainval_list, *, VALIDATION_SIZE=0.2, SEED=1):
    """
    separate a train list into a train list and a validation list
    :param train_list: list
    :param VALIDATION_SIZE: the size of the validation set (0.2 means 80 20 training validation respectively)
    :param SEED: seed for numpy
    :return: 2 lists, the first is the validation list, and the 2nd is the train list
    """

    middle = int(len(trainval_list) * VALIDATION_SIZE // 1)
    np.random.seed(SEED)
    np.random.shuffle(trainval_list)

    val_list = trainval_list[:middle]
    train_list = trainval_list[middle:]

    return val_list, train_list



import h5py
import os
import numpy as np

from .load_audio import load_audio


def create_hdf(
    hdf_dir_path,
    filename,
    data,
    *,
    source=None,
    output=["bass", "drums", "other", "vocals"],
    sr=44100,
    mono=False
):
    """
    Create an hdf database for the particular use case
    :param hdf_dir_path: path to hdf directory
    :param filename: the name of the hdf file (xxxx.hdf5)
    :param data: (list of dictionaries) the data to be inserted
    :param source: the key for the source audio path (key refers to the key in data), None would mean the source will be computed as the sum of output
    :param output: a list of keys to be used as the output
    :param sr: sample rate for the audio
    :param mono: True if data should be summed into mono
    :return: (String) path to the hdf5 file
    """
    if not os.path.exists(hdf_dir_path):
        os.makedirs(hdf_dir_path)

    hdf_path = os.path.join(hdf_dir_path, filename + ".hdf5")
    with h5py.File(hdf_path, "w") as f:
        # set f.attrs for sr, channels, and instruments maybe?

        for index, datum in enumerate(data):
            stem_audio = []
            for stem in output:
                audio_data, _ = load_audio(datum[stem], sr=sr, mono=mono)
                stem_audio.append(audio_data)

            if source:
                mix_data, _ = load_audio(datum[source], sr=sr, mono=mono)
            else:
                min_length = min([i.shape[1] for i in stem_audio])
                stem_audio = [i[:, :min_length] for i in stem_audio]
                mix_data = np.sum(stem_audio, axis=0)

            output_data = np.concatenate(stem_audio, axis=0)

            grp = f.create_group(str(index))
            grp.create_dataset(
                "source", shape=mix_data.shape, dtype=mix_data.dtype, data=mix_data
            )
            grp.create_dataset(
                "output",
                shape=output_data.shape,
                dtype=output_data.dtype,
                data=output_data,
            )

    return hdf_path
